
//-------------------------------------------------------------------------------//
// Author: Stefan Lörwald, Universität Heidelberg                                //
// License: CC BY-NC 4.0 http://creativecommons.org/licenses/by-nc/4.0/legalcode //
//-------------------------------------------------------------------------------//

#define COMPILE_TEMPLATE_ALGORITHM_INEQUALITY_OPERATIONS
#include "algorithm_inequality_operations.h"
#undef COMPILE_TEMPLATE_ALGORITHM_INEQUALITY_OPERATIONS

#include <algorithm>
#include <cassert>
#include <numeric>

namespace
{
   /// Calculates the scalar product of the left hand side coefficients of an inequality with a vertex.
   template <typename Integer>
   Integer scalarProduct(const Inequality<Integer>&, const Vertex<Integer>&);
}

template <typename Integer>
Integer rhs(const Inequality<Integer>& inequality)
{
   assert( !inequality.empty() );
   return -inequality.back();
}

template <typename Integer>
Integer distance(const Inequality<Integer>& inequality, const Vertex<Integer>& vertex)
{
   return -scalarProduct(inequality, vertex);
}

template <typename Integer>
Vertex<Integer> furthestVertex(const Vertices<Integer>& vertices, const Inequality<Integer>& inequality)
{
   assert( !vertices.empty() );
   Integer maximum = distance(inequality, *vertices.cbegin());
   auto best_vertex = vertices.cbegin();
   for ( auto it(vertices.cbegin() + 1); it != vertices.cend(); ++it )
   {
      const auto& vertex = *it;
      const auto d = distance(inequality, vertex);
      if ( d > maximum )
      {
         best_vertex = it;
         maximum = d;
      }
   }
   return *best_vertex;
}

template <typename Integer>
Vertex<Integer> nearestVertex(const Vertices<Integer>& vertices, const Inequality<Integer>& inequality)
{
   assert( !vertices.empty() );
   Integer minimum = distance(inequality, *vertices.cbegin());
   auto best_vertex = vertices.cbegin();
   for ( auto it(vertices.cbegin() + 1); it != vertices.cend(); ++it )
   {
      const auto& vertex = *it;
      const auto d = distance(inequality, vertex);
      if ( d < minimum )
      {
         best_vertex = it;
         minimum = d;
      }
   }
   return *best_vertex;
}

namespace
{
   template <typename Integer>
   Integer scalarProduct(const Inequality<Integer>& inequality, const Vertex<Integer>& vertex)
   {
      assert( inequality.size() == vertex.size() );
      return std::inner_product(vertex.cbegin(), vertex.cend(), inequality.cbegin(), Integer{0});
   }
}

