
//-------------------------------------------------------------------------------//
// Author: Stefan Lörwald, Universität Heidelberg                                //
// License: CC BY-NC 4.0 http://creativecommons.org/licenses/by-nc/4.0/legalcode //
//-------------------------------------------------------------------------------//

#define COMPILE_TEMPLATE_ALGORITHM_ROTATION
#include "algorithm_rotation.h"
#undef COMPILE_TEMPLATE_ALGORITHM_ROTATION

#include <algorithm>
#include <cassert>

#include "algorithm_facet_classes.h"
#include "algorithm_fourier_motzkin_elimination.h"
#include "algorithm_inequality_operations.h"
#include "algorithm_integer_operations.h"
#include "algorithm_row_operations.h"

namespace
{
   /// Rotates a facet around a ridge.
   template <typename Integer>
   Facet<Integer> rotate(const Vertices<Integer>&, const Facet<Integer>&, Facet<Integer>);
   /// Returns all ridges on a facet (equivalent to all facets of the facet).
   template <typename Integer>
   Inequalities<Integer> getRidges(const Vertices<Integer>&, const Facet<Integer>&, int = 0);
   /// Returns all vertices that lie on the facet (satisfy the inequality with equality).
   template <typename Integer>
   Vertices<Integer> verticesWithZeroDistance(const Vertices<Integer>&, const Facet<Integer>&);
}

template <typename Integer>
Facets<Integer> algorithm::rotation(const Vertices<Integer>& vertices,
                                    const Facet<Integer>& facet,
                                    const Maps& maps)
{
   const auto ridges = getRidges(vertices, facet);
   std::set<Facet<Integer>> facets;
   for ( const auto& ridge : ridges )
   {
      auto new_facet = rotate(vertices, facet, ridge);
      facets.insert(new_facet);
   }
   return facetClasses(facets, maps);
}

namespace
{
   template <typename Integer>
   Facet<Integer> rotate(const Vertices<Integer>& vertices, const Facet<Integer>& facet, Facet<Integer> ridge)
   {
      auto vertex = furthestVertex(vertices, facet); // possibly only some vertex with distance > 0
      auto d_f = distance(facet, vertex);
      auto d_r = distance(ridge, vertex);
      do
      {
         const auto gcd_ds = gcd(d_f, d_r);
         if ( gcd_ds > 1 )
         {
            d_f /= gcd_ds;
            d_r /= gcd_ds;
         }
         ridge = d_f * ridge - d_r * facet;
         const auto gcd_value = gcd(ridge);
         assert( gcd_value != 0 );
         if ( gcd_value > 1 )
         {
            ridge /= gcd_value;
         }
         vertex = nearestVertex(vertices, ridge);
         d_f = distance(facet, vertex);
         d_r = distance(ridge, vertex);
      }
      while ( d_r != 0 );
      return ridge;
   }

   template <typename Integer>
   Facets<Integer> fullRotationEnumeration(const Vertices<Integer>& vertices, int level)
   {
      const auto initial_facets = algorithm::fourierMotzkinEliminationHeuristic(vertices);
      std::set<Facet<Integer>> facets;
      for ( auto& facet : initial_facets )
      {
         facets.insert(facet);
      }
      using Iterator = typename std::set<Facet<Integer>>::iterator;
      std::vector<Iterator> iterators;
      for ( auto it(facets.cbegin()); it != facets.cend(); ++it )
      {
         iterators.push_back(it);
      }
      while ( !iterators.empty() )
      {
         const auto facet = *iterators.front();
         iterators.erase(iterators.begin());
         const auto ridges = getRidges(vertices, facet, level);
         for ( const auto& ridge : ridges )
         {
            auto new_facet = rotate(vertices, facet, ridge);
            Iterator iterator;
            bool inserted;
            std::tie(iterator, inserted) = facets.insert(new_facet);
            if ( inserted )
            {
               iterators.push_back(iterator);
            }
         }
      }
      return Facets<Integer>(facets.cbegin(), facets.cend());
   }

   template <typename Integer>
   Inequalities<Integer> getRidges(const Vertices<Integer>& vertices, const Facet<Integer>& facet, int level)
   {
      const auto vertices_on_facet = verticesWithZeroDistance(vertices, facet);
      assert( !vertices_on_facet.empty() );
      // EXPERIMENTAL
      //if ( vertices_on_facet.size() < 2 * vertices_on_facet.back().size() )
      if ( level == 0 )
      {
         return algorithm::fourierMotzkinElimination(vertices_on_facet);
      }
      else
      {
         return fullRotationEnumeration(vertices_on_facet, level + 1);
      }
   }

   template <typename Integer>
   Vertices<Integer> verticesWithZeroDistance(const Vertices<Integer>& vertices, const Facet<Integer>& facet)
   {
      Vertices<Integer> selection;
      for ( const auto& vertex : vertices )
      {
         if ( distance(facet, vertex) == 0 )
         {
            selection.push_back(vertex);
         }
      }
      return selection;
   }
}

