
//-------------------------------------------------------------------------------//
// Author: Stefan Lörwald, Universität Heidelberg                                //
// License: CC BY-NC 4.0 http://creativecommons.org/licenses/by-nc/4.0/legalcode //
//-------------------------------------------------------------------------------//

#include <array>
#include <cassert>
#include <limits>

#include "bitset_variable_size.h"
#include "popcount.h"

BitsetVariableSize::BitsetVariableSize(const std::size_t bits)
:
   data(1 + (bits - 1) / std::numeric_limits<DataType>::digits)
{
   assert( bits > 0 );
}

bool BitsetVariableSize::equals(const BitsetVariableSize& second, const std::size_t max) const
{
   assert( data.size() == second.data.size() );
   const auto end = data.cbegin() + 1 + (max - 1) / std::numeric_limits<DataType>::digits;
   return std::equal(data.cbegin(), end, second.data.cbegin());
}

bool BitsetVariableSize::contains(const BitsetVariableSize& second, const std::size_t max) const
{
   assert( data.size() == second.data.size() );
   const auto end = 1u + (max - 1) / std::numeric_limits<DataType>::digits;
   for ( std::size_t i(0); i < end; ++i )
   {
      if ( (data[i] & second.data[i]) != second.data[i] )
      {
         return false;
      }
   }
   return true;
}

std::size_t BitsetVariableSize::count(const std::size_t max) const
{
   std::size_t total{0};
   const auto end = data.cbegin() + 1 + (max - 1) / std::numeric_limits<DataType>::digits;
   for ( auto it(data.cbegin()); it != end; ++it )
   {
      total += popcount(*it);
   }
   return total;
}

BitsetVariableSize BitsetVariableSize::merge(const BitsetVariableSize& second, const std::size_t max) const
{
   assert( data.size() == second.data.size() );
   BitsetVariableSize result = *this;
   const std::size_t end = 1 + (max - 1) / std::numeric_limits<DataType>::digits;
   for ( std::size_t i(0); i < end; ++i )
   {
      result.data[i] |= second.data[i];
   }
   return result;
}

void BitsetVariableSize::set(const std::size_t index)
{
   const DataType mask = static_cast<DataType>(1u) << (index % std::numeric_limits<DataType>::digits);
   assert(index / std::numeric_limits<DataType>::digits < data.size() );
   data[index / std::numeric_limits<DataType>::digits] |= mask;
}

