
//-------------------------------------------------------------------------------//
// Author: Stefan Lörwald, Universität Heidelberg                                //
// License: CC BY-NC 4.0 http://creativecommons.org/licenses/by-nc/4.0/legalcode //
//-------------------------------------------------------------------------------//

#pragma once

#include <chrono>

template <typename Runnable>
class DelayedAction;

/// Creates a delayed action that runs if at destruction the elapsed time since
/// creation of the delayed action exceeds the specified amount of seconds.
template <typename Runnable>
DelayedAction<Runnable> makeDelayedAction(Runnable&&, std::chrono::seconds);
/// Creates a delayed action that runs if at destruction the elapsed time since
/// creation of the delayed action exceeds the specified amount of milliseconds.
template <typename Runnable>
DelayedAction<Runnable> makeDelayedAction(Runnable&&, std::chrono::milliseconds);
/// Creates a delayed action that runs if at destruction the elapsed time since
/// creation of the delayed action exceeds the specified amount of nanoseconds.
template <typename Runnable>
DelayedAction<Runnable> makeDelayedAction(Runnable&&, std::chrono::nanoseconds);

template <typename Runnable>
class DelayedAction
{
   public:
      friend DelayedAction<Runnable> makeDelayedAction<>(Runnable&&, std::chrono::nanoseconds);
   private:
      /// Constructor.
      DelayedAction(Runnable&&, std::chrono::nanoseconds);
   public:
      /// Destructor.
      ~DelayedAction();
      /// If the delayed action is dismissed, the action will not be performed
      /// no matter how the scope is exited.
      void dismiss();
   private:
      bool dismissed;
      Runnable runnable;
      std::chrono::time_point<std::chrono::system_clock> start;
      std::chrono::nanoseconds minimum_waiting_time;
};

#include "delayed_action.tpp"

