
//-------------------------------------------------------------------------------//
// Author: Stefan Lörwald, Universität Heidelberg                                //
// License: CC BY-NC 4.0 http://creativecommons.org/licenses/by-nc/4.0/legalcode //
//-------------------------------------------------------------------------------//

#include "input.h"

#include <algorithm>
#include <cassert>
#include <cstring>
#include <fstream>
#include <limits>
#include <stdexcept>
#include <string>

#include "input_common.h"
#include "input_detection.h"
#include "input_dimension.h"
#include "input_inequality.h"
#include "input_keywords.h"
#include "input_map.h"
#include "input_names.h"
#include "input_order.h"
#include "input_validity.h"
#include "input_vertex.h"
#include "istream_peek_line.h"

namespace
{
   void sort(int, char**, Matrix<int>&);
}

template <>
std::tuple<Vertices<int>, Names, Maps> input::vertices<int>(int argc, char** argv)
{
   /// In this method, we expect a file containing a conv/cone description in PANDA format.
   /// Raw format is not supported as deduction of which one is the convex hull part is
   /// impossible from data. We additionally may have names and maps.
   const auto filename = getFilename(argc, argv);
   std::ifstream file(filename.c_str());
   if ( !file )
   {
      throw std::invalid_argument("Failed to open file \"" + filename + "\".");
   }
   ConvexHull<int> conv;
   ConvexHull<int> cone;
   Names names;
   Maps maps;
   std::size_t dimension = std::numeric_limits<std::size_t>::max();
   for ( std::string token; advanceToNextKeyword(file, token); )
   {
      if ( implementation::isKeywordDimension(token) )
      {
         dimension = implementation::dimension(file);
      }
      else if ( implementation::isKeywordConvexHull(token) )
      {
         conv = implementation::verticesConvex(file);
      }
      else if ( implementation::isKeywordConicalHull(token) )
      {
         cone = implementation::verticesConical(file);
      }
      else if ( implementation::isKeywordNames(token) )
      {
         names = implementation::names(file);
      }
      else if ( implementation::isKeywordMaps(token) )
      {
         maps = implementation::maps(file, names);
      }
      else
      {
         std::getline(file, token);
      }
   }
   implementation::checkValidity(conv, cone, names, maps, dimension);
   sort(argc, argv, cone);
   sort(argc, argv, conv);
   Matrix<int> vertices;
   vertices.reserve(cone.size() + conv.size());
   vertices.insert(vertices.end(), cone.cbegin(), cone.cend());
   vertices.insert(vertices.end(), conv.cbegin(), conv.cend());
   return std::make_tuple(vertices, names, maps);
}

template <>
Inequalities<int> input::inequalities<int>(int argc, char** argv)
{
   /// In this method we expect a file containing an outer description
   /// and a valid vertex in PANDA format or raw format.
   const auto filename = getFilename(argc, argv);
   std::ifstream file(filename.c_str());
   if ( !file )
   {
      throw std::invalid_argument("Failed to open file \"" + filename + "\".");
   }
   Inequalities<int> inequalities;
   std::size_t dimension = std::numeric_limits<std::size_t>::max();
   Names names;
   if ( !implementation::containsKeywords(file) )
   {
      throw std::invalid_argument("An outer description must be given in PANDA format.");
   }
   for ( std::string token; file && advanceToNextKeyword(file, token); )
   {
      if ( implementation::isKeywordDimension(token) )
      {
         dimension = implementation::dimension(file);
      }
      else if ( implementation::isKeywordNames(token) )
      {
         names = implementation::names(file);
      }
      else if ( implementation::isKeywordInequalities(token) )
      {
         inequalities = implementation::inequalitiesPanda(file, names);
      }
      else
      {
         std::getline(file, token);
      }
   }
   implementation::checkValidity(inequalities, names, dimension);
   sort(argc, argv, inequalities);
   return inequalities;
}

namespace
{
   std::string getFilenameKnownFacets(int argc, char** argv)
   {
      for ( int i(1); i < argc; ++i )
      {
         if ( std::strcmp(argv[i], "-k") == 0 )
         {
            if ( i + 1 == argc )
            {
               throw std::invalid_argument("Expected argument to option \"--known-facets\".");
            }
            return argv[i + 1];
         }
         if ( std::strncmp(argv[i], "--known-facets=", 15) == 0 )
         {
            return argv[i] + 15;
         }
      }
      return "";
   }
}

template <>
Inequalities<int> input::knownFacets<int>(int argc, char** argv, const Names& names)
{
   const auto filename = getFilenameKnownFacets(argc, argv);
   if ( filename.empty() )
   {
      return {};
   }
   std::ifstream file(filename.c_str());
   if ( !file )
   {
      throw std::invalid_argument("Failed to open file \"" + filename + "\".");
   }
   Inequalities<int> inequalities;
   for ( std::string token; file && advanceToNextKeyword(file, token); )
   {
      inequalities = implementation::inequalitiesPanda(file, names);
   }
   return inequalities;
}

template <>
Polytope<int> input::polytope<int>(int argc, char** argv)
{
   /// In this method we expect a file containing an outer description
   /// and a conv/cone description, both in PANDA format.
   /// We additionally may have names and maps.
   ConvexHull<int> conv;
   ConicalHull<int> cone;
   Inequalities<int> inequalities;
   Names names;
   Maps maps;
   std::size_t dimension = std::numeric_limits<std::size_t>::max();
   const auto filename = getFilename(argc, argv);
   std::ifstream file(filename.c_str());
   if ( !file )
   {
      throw std::invalid_argument("Failed to open file \"" + filename + "\".");
   }
   for ( std::string token; advanceToNextKeyword(file, token); )
   {
      if ( implementation::isKeywordDimension(token) )
      {
         dimension = implementation::dimension(file);
      }
      else if ( implementation::isKeywordConvexHull(token) )
      {
         conv = implementation::verticesConvex(file);
      }
      else if ( implementation::isKeywordConicalHull(token) )
      {
         cone = implementation::verticesConical(file);
      }
      else if ( implementation::isKeywordNames(token) )
      {
         names = implementation::names(file);
      }
      else if ( implementation::isKeywordInequalities(token) )
      {
         inequalities = implementation::inequalitiesPanda(file, names);
      }
      else if ( implementation::isKeywordMaps(token) )
      {
         maps = implementation::maps(file, names);
      }
      else
      {
         std::getline(file, token); // consume ignored token
      }
   }
   implementation::checkValidity(conv, cone, inequalities, names, maps, dimension);
   return std::make_tuple(conv, cone, inequalities, names, maps);
}

#include "algorithm_row_operations.h"

namespace
{
   void sort(int argc, char** argv, Matrix<int>& matrix)
   {
      const auto order = getInputOrder(argc, argv);
      switch ( order )
      {
         case InputOrder::LexicographicAscending:
         {
            std::sort(matrix.begin(), matrix.end());
            return;
         }
         case InputOrder::LexicographicDescending:
         {
            std::sort(matrix.rbegin(), matrix.rend());
            return;
         }
         case InputOrder::NonZeroEntriesAscending:
         {
            std::sort(matrix.begin(), matrix.end(), [](const Row<int>& a, const Row<int>& b)
            {
               const auto zero_entries_a = std::count(a.cbegin(), a.cend(), 0);
               const auto zero_entries_b = std::count(b.cbegin(), b.cend(), 0);
               if ( zero_entries_a != zero_entries_b )
               {
                  return (zero_entries_a > zero_entries_b);
               }
               return (a < b);
            });
            return;
         }
         case InputOrder::NonZeroEntriesDescending:
         {
            std::sort(matrix.begin(), matrix.end(), [](const Row<int>& a, const Row<int>& b)
            {
               const auto zero_entries_a = std::count(a.cbegin(), a.cend(), 0);
               const auto zero_entries_b = std::count(b.cbegin(), b.cend(), 0);
               if ( zero_entries_a != zero_entries_b )
               {
                  return (zero_entries_a < zero_entries_b);
               }
               return (a < b);
            });
            return;
         }
         case InputOrder::Reverse:
         {
            std::reverse(matrix.begin(), matrix.end());
            return;
         }
         case InputOrder::NoSorting:
         {
            return;
         }
         default:
         {
            throw std::invalid_argument("Invalid sorting command. (This is most likely a bug!)");
         }
      }
   }
}

