
//-------------------------------------------------------------------------------//
// Author: Stefan Lörwald, Universität Heidelberg                                //
// License: CC BY-NC 4.0 http://creativecommons.org/licenses/by-nc/4.0/legalcode //
//-------------------------------------------------------------------------------//

#include "input_common.h"

#include <algorithm>
#include <sstream>
#include <stdexcept>

#include "algorithm_row_operations.h"
#include "input_keywords.h"
#include "istream_peek_line.h"
#include "scope_guard.h"

bool input::isIdentifier(std::istream& stream)
{
   const auto original_position = stream.tellg();
   const auto guard = makeScopeGuard([&stream, original_position]()
   {
      stream.clear();
      stream.seekg(original_position);
   });
   return stream && stream.peek() == '(';
}

std::string input::identifier(std::istream& stream)
{
   std::string string;
   if ( stream.peek() != '(' )
   {
      throw std::invalid_argument("Identifier must start with '('.");
   }
   stream.get(); // consume the `(`.
   skipWhitespace(stream);
   while ( stream &&
           stream.peek() != '\n' &&
           stream.peek() != std::istream::traits_type::eof() &&
           stream.peek() != ')' )
   {
      string.push_back(stream.get());
   }
   if ( stream.peek() != ')' )
   {
      std::string message = "Identifier must end with a ')'.";
      message += " The character `)` and the newline character `\\n`";
      message += " may not be contained in the identifier.";
      throw std::invalid_argument(message);
   }
   stream.get(); // consume the `)`.
   const auto start_of_whitespace = string.find_last_not_of(" \t");
   if ( start_of_whitespace != std::string::npos )
   {
      string.erase(start_of_whitespace + 1);
   }
   return string;
}

bool input::isRelation(std::istream& stream)
{
   const auto original_position = stream.tellg();
   const auto guard = makeScopeGuard([&stream, original_position]()
   {
      stream.clear();
      stream.seekg(original_position);
   });
   if ( !stream )
   {
      return false;
   }
   const auto c = stream.peek();
   if ( c == '<' || c == '=' || c == '>' )
   {
      stream.get();
      if ( (stream && stream.peek() == '=') || !stream || isWhitespace(stream) )
      {
         return true;
      }
   }
   return false;
}

input::RelationOperator input::relationOperator(std::istream& stream)
{
   if ( !stream )
   {
      throw std::invalid_argument("Not a relation operator.");
   }
   const auto first = stream.get();
   if ( first == '<' )
   {
      if ( stream && stream.peek() == '=' )
      {
         stream.get();
         return RelationOperator::LessEqual;
      }
      else if ( !stream || isWhitespace(stream) )
      {
         return RelationOperator::Less;
      }
      else
      {
         throw std::invalid_argument("Relation operator \"<\" must be followed by whitespace.");
      }
   }
   else if ( first == '=' )
   {
      if ( stream && stream.peek() == '=' )
      {
         stream.get();
         return RelationOperator::Equal;
      }
      else if ( !stream || isWhitespace(stream) )
      {
         return RelationOperator::Equal;
      }
      else
      {
         throw std::invalid_argument("Relation operator \"=\" must be followed by whitespace.");
      }
   }
   else if ( first == '>' )
   {
      if ( stream && stream.peek() == '=' )
      {
         stream.get();
         return RelationOperator::GreaterEqual;
      }
      else if ( !stream || isWhitespace(stream) )
      {
         return RelationOperator::Greater;
      }
      else
      {
         throw std::invalid_argument("Relation operator \">\" must be followed by whitespace.");
      }
   }
   throw std::invalid_argument("Not a relation operator.");
}

bool input::isWhitespace(std::istream& stream)
{
   const auto original_position = stream.tellg();
   const auto guard = makeScopeGuard([&stream, original_position]()
   {
      stream.clear();
      stream.seekg(original_position);
   });
   if ( !stream )
   {
      return false;
   }
   const auto c = stream.peek();
   return c == ' ' || c == '\t';
}

void input::skipWhitespace(std::istream& stream)
{
   while ( stream && isWhitespace(stream) )
   {
      stream.get();
   }
}

bool input::containsNames(const Names& names, std::istream& stream)
{
   const auto stream_position = stream.tellg();
   const auto guard = makeScopeGuard([&] ()
   {
      stream.clear();
      stream.seekg(stream_position);
   });
   std::size_t counter = 0;
   constexpr std::size_t max_counter = 2000;
   while ( stream && counter < max_counter )
   {
      auto position = names.cbegin();
      std::string word = "";
      while ( stream && position != names.cend() )
      {
         const auto character = stream.get();
         word.append(1, character);
         position = std::find_if(names.cbegin(), names.cend(), [&] (const Name& name) { return word.size() <= name.size() && name.substr(0, word.size()) == word; });
         if ( position != names.cend() && *position == word ) // match!
         {
            return true;
         }
         ++counter;
      }
   }
   // assuming that if a name occurs it will do so in the first few characters
   return false;
}

Row<int> input::normalize(Row<int> row, RelationOperator relation)
{
   // invert the inequality if relation is Greater or GreaterEqual
   if ( relation == RelationOperator::GreaterEqual || relation == RelationOperator::Greater )
   {
      row *= -1; // invert, now row satisfies LessEqual / Less.
   }
   // from now on, row satisfies Less, LessEqual or Equal, although relation may still say something else.
   if ( relation == RelationOperator::Less || relation == RelationOperator::Greater )
   {                   // rhs can be decreased by one
      row.back() += 1; // rhs is inverted, hence decreased rhs is increased -rhs.
   }
   return row;
}

int input::toInt(const std::string& string)
{
   if ( string.empty() || string == "+" )
   {
      return 1;
   }
   if ( string == "-" )
   {
      return -1;
   }
   std::istringstream stream(string);
   int result;
   if ( !(stream >> result) )
   {
      throw std::invalid_argument("toInt: Argument is not an integer.");
   }
   return result;
}

std::istream& input::advanceToNextKeyword(std::istream& stream, std::string& string)
{
   while ( stream )
   {
      string = peekLine(stream);
      if ( implementation::isKeyword(string) )
      {
         break;
      }
      using std::getline;
      getline(stream, string);
   }
   return stream;
}

std::string input::firstWord(std::string line)
{
   const auto beg_word = line.find_first_not_of(" \t");
   line.erase(0, beg_word);
   const auto end_word = line.find_first_of(" \t");
   if ( end_word != std::string::npos )
   {
      line.erase(end_word);
   }
   return line;
}

std::string input::peekWord(std::istream& stream)
{
   const auto original_position = stream.tellg();
   auto guard = makeScopeGuard([&stream, original_position]()
   {
      stream.clear();
      stream.seekg(original_position);
   });
   std::string token;
   if ( !(stream >> token) )
   {
      guard.dismiss();
   }
   return token;
}

