
//-------------------------------------------------------------------------------//
// Author: Stefan Lörwald, Universität Heidelberg                                //
// License: CC BY-NC 4.0 http://creativecommons.org/licenses/by-nc/4.0/legalcode //
//-------------------------------------------------------------------------------//

#include "input_inequality_unnamed_variables.h"

#include <stdexcept>

#include "input_common.h"

/// We have to deal with two possible input types here:
/// Raw format (no names, no relation operator) and PANDA format without names.
/// This only differs in the presence of the missing relation operator.
/// We allow a leading "()"-enclosed identifier in both cases.
std::pair<Inequality<int>, ConstraintType> input::implementation::inequality(std::istream& stream)
{
   Inequality<int> ieq;
   RelationOperator relation = RelationOperator::Undecided;
   bool has_rhs = false;
   skipWhitespace(stream);
   const auto id = isIdentifier(stream) ? identifier(stream) : "";
   skipWhitespace(stream);
   while ( stream && stream.peek() != '\n' && stream.peek() != std::istream::traits_type::eof() )
   {
      if ( isRelation(stream) )
      {
         relation = relationOperator(stream);
      }
      else
      {
         int coefficient;
         if ( !(stream >> coefficient) )
         {
            error<std::invalid_argument>("Non-integral token in inequality", id);
         }
         if ( has_rhs )
         {
            error<std::invalid_argument>("Only one coefficient allowed on right hand side", id);
         }
         if ( relation != RelationOperator::Undecided )
         {
            has_rhs = true;
         }
         ieq.push_back(coefficient);
      }
      skipWhitespace(stream);
   }
   if ( relation != RelationOperator::Undecided && !has_rhs )
   {
      error<std::invalid_argument>("No right hand side provided", id);
   }
   if ( ieq.empty() )
   {
      return std::make_pair(ieq, ConstraintType::Equality);
   }
   if ( relation != RelationOperator::Undecided ) // invert the last entry if input was in PANDA format
   {
      ieq.back() *= -1;
   }
   const auto type = (relation == RelationOperator::Equal) ? ConstraintType::Equality : ConstraintType::Inequality;
   return std::make_pair(normalize(ieq, relation), type);
}

