
//-------------------------------------------------------------------------------//
// Author: Stefan Lörwald, Universität Heidelberg                                //
// License: CC BY-NC 4.0 http://creativecommons.org/licenses/by-nc/4.0/legalcode //
//-------------------------------------------------------------------------------//

#include "input_vertex.h"

#include <cassert>
#include <stdexcept>
#include <string>

#include "algorithm_row_operations.h"
#include "input_common.h"
#include "input_keywords.h"
#include "istream_peek_line.h"

namespace
{
   Row<int> vertex(std::istream&);
}

ConicalHull<int> input::implementation::verticesConical(std::istream& stream)
{
   ConicalHull<int> cone;
   std::string token;
   if ( !(stream >> token) || !isKeywordConicalHull(token) )
   {
      throw std::invalid_argument("Cannot read conical hull: file is at an invalid position.");
   }
   while ( stream && input::firstWord(peekLine(stream)) != "END" )
   {
      auto ray = ::vertex(stream);
      if ( !ray.empty() )
      {
         ray.back() = 0;
         cone.push_back(ray);
      }
      else
      {
         stream.clear();
      }
   }
   if ( stream )
   {
      using std::getline;
      getline(stream, token); // consume "END".
   }
   return cone;
}

ConvexHull<int> input::implementation::verticesConvex(std::istream& stream)
{
   ConvexHull<int> conv;
   std::string token;
   if ( !(stream >> token) || !isKeywordConvexHull(token) )
   {
      throw std::invalid_argument("Cannot read convex hull: file is at an invalid position.");
   }
   while ( stream && input::firstWord(peekLine(stream)) != "END" )
   {
      auto line = peekLine(stream);
      line.erase(0, line.find_first_not_of(" \t"));
      if ( line.empty() )
      {
         using std::getline;
         getline(stream, token);
         continue;
      }
      auto vertex = ::vertex(stream);
      if ( !vertex.empty() )
      {
         conv.push_back(vertex);
      }
      else
      {
         throw std::invalid_argument("Invalid line content in convex hull: \"" + line + "\".");
      }
   }
   if ( stream )
   {
      using std::getline;
      getline(stream, token); // consume "END".
   }
   return conv;
}

ConvexHull<int> input::implementation::vertices(std::istream& stream)
{
   ConvexHull<int> conv;
   while ( stream )
   {
      const auto vertex = ::vertex(stream);
      if ( !vertex.empty() )
      {
         conv.push_back(vertex);
      }
   }
   return conv;
}

namespace
{
   Row<int> vertex(std::istream& stream)
   {
      input::skipWhitespace(stream);
      const auto id = input::isIdentifier(stream) ? input::identifier(stream) : "";
      input::skipWhitespace(stream);
      Row<int> numerator;
      Row<int> denominator;
      for ( int tmp; stream >> tmp; )
      {
         numerator.push_back(tmp);
         input::skipWhitespace(stream);
         if ( stream.peek() == '/' )
         {
            stream.get(); // consume division operator
            stream >> tmp;
            if ( tmp == 0 )
            {
               input::error<std::invalid_argument>("Division by zero.", id);
            }
            denominator.push_back(tmp);
         }
         else
         {
            denominator.push_back(1);
         }
         while ( stream && stream.peek() != std::istream::traits_type::eof() && (stream.peek() == ' ' || stream.peek() == '\t') )
         {
            stream.get();
         }
         if ( stream.peek() == '\n' )
         {
            break;
         }
      }
      assert( numerator.size() == denominator.size() );
      if ( numerator.empty() )
      {
         return numerator;
      }
      const auto lcm_value = lcm(denominator);
      if ( lcm_value > 0 )
      {
         for ( std::size_t i(0); i < numerator.size(); ++i )
         {
            numerator[i] *= lcm_value / denominator[i];
         }
      }
      numerator.push_back(lcm_value);
      return numerator;
   }
}

