
//-------------------------------------------------------------------------------//
// Author: Stefan Lörwald, Universität Heidelberg                                //
// License: CC BY-NC 4.0 http://creativecommons.org/licenses/by-nc/4.0/legalcode //
//-------------------------------------------------------------------------------//

#include <cstdint>
#include <iostream>
#include <stdexcept>

#include "big_integer.h"
#include "integer_type_detection.h"
#include "safe_integer.h"

template <template <typename> class Functor>
int IntegerTypeSelector<Functor>::select(int argc, char** argv)
{
   const auto integer_type = integerType(argc, argv);
   switch ( integer_type )
   {
      #ifndef NO_FLEXIBILITY
      case IntegerType::Fixed16:
      {
         #ifdef INT16_MIN
         return Functor<int16_t>::call(argc, argv);
         #else
         std::cerr << "On this machine, there is no integer type which is exactly 16 bits wide.\n"
                   << "Calculations are performed with default integer type.\n";
         #endif
         goto default_type;
      }
      case IntegerType::Fixed32:
      {
         #ifdef INT32_MIN
         return Functor<int32_t>::call(argc, argv);
         #else
         std::cerr << "On this machine, there is no integer type which is exactly 32 bits wide.\n"
                   << "Calculations are performed with default integer type.\n";
         #endif
         goto default_type;
      }
      case IntegerType::Fixed64:
      {
         #ifdef INT64_MIN
         return Functor<int64_t>::call(argc, argv);
         #else
         std::cerr << "On this machine, there is no integer type which is exactly 64 bits wide.\n"
                   << "Calculations are performed with default integer type.\n";
         #endif
         goto default_type;
      }
      case IntegerType::Safe:
      {
         return Functor<SafeInteger>::call(argc, argv);
      }
      case IntegerType::Variable:
      {
         return Functor<BigInteger>::call(argc, argv);
      }
      #else
      default:
      {
         std::cerr << "You compiled with \"-DNO_FLEXIBILITY\"."
                   << " Selection of fixed size integer type or arbitrary precision integer is disabled."
                   << " Calculations will be done with the fastest integer type with at least 32 bits.\n";
         goto default_type;
      }
      #endif
      case IntegerType::Default:
      {
         default_type:
         return Functor<int>::call(argc, argv);
      }
   }
   throw std::invalid_argument("Developer made a mistake handling IntegerType");
}

