
//-------------------------------------------------------------------------------//
// Author: Stefan Lörwald, Universität Heidelberg                                //
// License: CC BY-NC 4.0 http://creativecommons.org/licenses/by-nc/4.0/legalcode //
//-------------------------------------------------------------------------------//

#include "method_adjacency_decomposition.h"

#include <iostream>
#include <exception>

#include "application_name.h"
#include "concurrency.h"
#include "delayed_action.h"
#include "input.h"
#include "integer_type_selection.h"
#include "job_manager.h"
#include "job_manager_proxy.h"
#include "message_passing_interface_session.h"
#include "method_adjacency_decomposition_implementation.h"

namespace
{
   /// Implementation of facet enumeration with adjacency decomposition.
   template <typename Integer>
   struct AdjacencyDecompositionImplementationFunctor
   {
      static int call(int, char**);
   };
}

int method::adjacencyDecomposition(int argc, char** argv)
{
   const auto& mpi_session = getMessagePassingInterfaceSession();
   if ( mpi_session.isMaster() )
   {
      std::cerr << APP_NAME << " -- facet enumeration with adjacency decomposition\n";
   }
   return IntegerTypeSelector<AdjacencyDecompositionImplementationFunctor>::select(argc, argv);
}

namespace
{
   template <typename Integer>
   int AdjacencyDecompositionImplementationFunctor<Integer>::call(int argc, char** argv)
   {
      std::tuple<Vertices<Integer>, Names, Maps> data;
      Inequalities<Integer> known_facets;
      try
      {
         const auto action = makeDelayedAction([](){ std::cerr << "Input completed\n"; },
                                               std::chrono::seconds(2));
         data = input::vertices<Integer>(argc, argv);
         known_facets = input::knownFacets<Integer>(argc, argv, std::get<1>(data));
      }
      catch ( const std::exception& e )
      {
         std::cerr << "Exception caught: " << e.what() << "\n";
         return 1;
      }
      catch ( ... )
      {
         std::cerr << "Unknown exception caught while reading files\n";
         return 1;
      }
      try
      {
         const auto& mpi_session = getMessagePassingInterfaceSession();
         const auto node_count = mpi_session.getNumberOfNodes();
         const auto thread_count = numberOfThreads(argc, argv);
         if ( mpi_session.isMaster() )
         {
            implementation::adjacencyDecomposition<JobManager>(data, known_facets, thread_count, node_count);
         }
         else
         {
            implementation::adjacencyDecomposition<JobManagerProxy>(data, known_facets, thread_count, node_count);
         }
         return 0;
      }
      catch ( const std::exception& e )
      {
         std::cerr << "Exception caught: " << e.what() << "\n";
         return 1;
      }
      catch ( ... )
      {
         std::cerr << "Unknown exception caught while performing facet enumeration"
                   << " with adjacency decomposition.\n";
         return 1;
      }
   }
}

