
//-------------------------------------------------------------------------------//
// Author: Stefan Lörwald, Universität Heidelberg                                //
// License: CC BY-NC 4.0 http://creativecommons.org/licenses/by-nc/4.0/legalcode //
//-------------------------------------------------------------------------------//

#include "method_double_description.h"

#include <exception>
#include <iostream>
#include <utility>
#include <set>
#include <tuple>

#include "algorithm_facet_classes.h"
#include "algorithm_fourier_motzkin_elimination.h"
#include "algorithm_map_operations.h"
#include "algorithm_matrix_operations.h"
#include "application_name.h"
#include "delayed_action.h"
#include "input.h"
#include "integer_type_selection.h"
#include "matrix.h"

namespace
{
   /// Implementation of facet enumeration with double description method.
   template <typename Integer>
   struct DoubleDescriptionImplementationFunctor
   {
      static int call(int, char**);
   };
}

int method::doubleDescription(int argc, char** argv)
{
   std::cerr << APP_NAME << " -- facet enumeration with double description method\n";
   return IntegerTypeSelector<DoubleDescriptionImplementationFunctor>::select(argc, argv);
}

namespace
{
   template <typename Integer>
   int DoubleDescriptionImplementationFunctor<Integer>::call(int argc, char** argv)
   {
      std::tuple<Vertices<Integer>, Names, Maps> data;
      try
      {
         const auto action = makeDelayedAction([](){ std::cerr << "Input completed\n"; },
                                               std::chrono::seconds(2));
         data = input::vertices<Integer>(argc, argv);
      }
      catch ( const std::exception& e )
      {
         std::cerr << "Exception caught: " << e.what() << "\n";
         return 1;
      }
      catch ( ... )
      {
         std::cerr << "Unknown exception caught while reading files\n";
         return 1;
      }
      try
      {
         const auto& vertices = std::get<0>(data);
         const auto& names = std::get<1>(data);
         auto& maps = std::get<2>(data);
         const auto equations = algorithm::extractEquations(vertices);
         if ( !equations.empty() )
         {
            std::cout << "Equations:\n";
            prettyPrint(std::cout, equations, names, "=");
            std::cout << "\n";
         }
         auto inequalities = algorithm::fourierMotzkinElimination(vertices);
         if ( !maps.empty() )
         {
            maps = algorithm::normalize(maps, equations);
            inequalities = algorithm::facetClasses(inequalities, maps);
         }
         std::cout << "Inequalities:\n";
         prettyPrint(std::cout, inequalities, names, "<=");
         return 0;
      }
      catch ( const std::exception& e )
      {
         std::cerr << "Exception caught: " << e.what() << "\n";
         return 1;
      }
      catch ( ... )
      {
         std::cerr << "Unknown exception caught in facet enumeration"
                   << " with double description method\n";
         return 1;
      }
   }
}

