
//-------------------------------------------------------------------------------//
// Author: Stefan Lörwald, Universität Heidelberg                                //
// License: CC BY-NC 4.0 http://creativecommons.org/licenses/by-nc/4.0/legalcode //
//-------------------------------------------------------------------------------//

#include "method_vertex_enumeration.h"

#include <cassert>
#include <exception>
#include <iostream>
#include <utility>
#include <tuple>

#include "algorithm_fourier_motzkin_elimination.h"
#include "algorithm_matrix_operations.h"
#include "algorithm_row_operations.h"
#include "application_name.h"
#include "delayed_action.h"
#include "input.h"
#include "integer_type_selection.h"
#include "matrix.h"
#include "row.h"

namespace
{
   /// Implementation of vertex enumeration.
   template <typename Integer>
   struct VertexEnumerationImplementationFunctor
   {
      static int call(int, char**);
   };
}

int method::vertexEnumeration(int argc, char** argv)
{
   std::cerr << APP_NAME << " -- vertex enumeration\n";
   return IntegerTypeSelector<VertexEnumerationImplementationFunctor>::select(argc, argv);
}

namespace
{
   template <typename Integer>
   int VertexEnumerationImplementationFunctor<Integer>::call(int argc, char** argv)
   {
      Inequalities<Integer> inequalities;
      try
      {
         const auto action = makeDelayedAction([](){ std::cerr << "Input completed\n"; },
                                               std::chrono::seconds(2));
         inequalities = input::inequalities<Integer>(argc, argv);
         assert( !inequalities.empty() );
         inequalities.push_back(Row<Integer>(inequalities.back().size(), Integer(0)));
         inequalities.back().back() = Integer(-1);
      }
      catch ( const std::exception& e )
      {
         std::cerr << "Exception caught: " << e.what() << "\n";
         return 1;
      }
      catch ( ... )
      {
         std::cerr << "Unknown exception caught while reading files\n";
         return 1;
      }
      try
      {
         auto vertices = algorithm::fourierMotzkinElimination(inequalities);
         std::vector<Vertex<Integer>*> conv;
         std::vector<Vertex<Integer>*> cone;
         for ( auto& row : vertices )
         {
            if ( row.back() != 0 )
            {
               conv.push_back(&row);
            }
            else
            {
               cone.push_back(&row);
            }
         }
         if ( !conv.empty() )
         {
            std::cout << "CONV_SECTION\n";
            for ( auto vertex : conv )
            {
               printFractional(std::cout, *vertex);
               std::cout << '\n';
            }
            std::cout << "END\n";
         }
         if ( !cone.empty() )
         {
            std::cout << "CONE_SECTION\n";
            for ( auto ray : cone )
            {
               ray->pop_back();
               std::cout << *ray << '\n';
            }
            std::cout << "END\n";
         }
         return 0;
      }
      catch ( const std::exception& e )
      {
         std::cerr << "Exception caught: " << e.what() << "\n";
         return 1;
      }
      catch ( ... )
      {
         std::cerr << "Unknown exception caught while enumerating vertices\n";
         return 1;
      }
   }
}

