
//-------------------------------------------------------------------------------//
// Author: Stefan Lörwald, Universität Heidelberg                                //
// License: CC BY-NC 4.0 http://creativecommons.org/licenses/by-nc/4.0/legalcode //
//-------------------------------------------------------------------------------//

#include "testing_gear.h"

#include "safe_integer.h"

#include <limits>

namespace
{
   void test_operator_multiply_assign();
   void test_operator_divide_assign();
   void test_operator_add_assign();
   void test_operator_subtract_assign();
   void test_operator_modulo_assign();
   void test_operator_multiply();
   void test_operator_divide();
   void test_operator_add();
   void test_operator_subtract();
   void test_operator_modulo();
}

int main()
try
{
   test_operator_multiply_assign();
   test_operator_divide_assign();
   test_operator_add_assign();
   test_operator_subtract_assign();
   test_operator_modulo_assign();
   test_operator_multiply();
   test_operator_divide();
   test_operator_add();
   test_operator_subtract();
   test_operator_modulo();
}
catch ( const std::exception& e )
{
   std::cerr << e.what() << "\n";
   return 1;
}

namespace
{
   using SI = SafeInteger;
   using SID = SI::DataType;
   using SIDL = std::numeric_limits<SID>;

   void test_operator_multiply_assign()
   {
      ASSERT(((SI(2) *= SI(3)) == SI(6)), "operator*=(SafeInteger)");
      ASSERT(((SI(2) *= SI(-3)) == SI(-6)), "operator*=(SafeInteger)");
      ASSERT(((SI(-2) *= SI(3)) == SI(-6)), "operator*=(SafeInteger)");
      ASSERT(((SI(-2) *= SI(-3)) == SI(6)), "operator*=(SafeInteger)");
      ASSERT(((SI(0) *= SI(3)) == SI(0)), "operator*=(SafeInteger)");
      ASSERT(((SI(2) *= SI(0)) == SI(0)), "operator*=(SafeInteger)");
      ASSERT_NOTHROW(SI(SIDL::max()) *= SI(1), "operator*=(SafeInteger)")
      ASSERT_ANY_EXCEPTION(SI(SIDL::max()) *= SI(2), "operator*=(SafeInteger)")
      ASSERT_NOTHROW(SI(SIDL::min()) *= SI(1), "operator*=(SafeInteger)")
      ASSERT_ANY_EXCEPTION(SI(SIDL::min()) *= SI(2), "operator*=(SafeInteger)")
   }

   void test_operator_divide_assign()
   {
      ASSERT(((SI(6) /= SI(3)) == SI(2)), "operator/=(SafeInteger)");
      ASSERT(((SI(6) /= SI(-3)) == SI(-2)), "operator/=(SafeInteger)");
      ASSERT(((SI(-6) /= SI(3)) == SI(-2)), "operator/=(SafeInteger)");
      ASSERT(((SI(-6) /= SI(-3)) == SI(2)), "operator/=(SafeInteger)");
      ASSERT(((SI(0) /= SI(3)) == SI(0)), "operator/=(SafeInteger)");
      ASSERT_ANY_EXCEPTION(((SI(6) /= SI(0)) == SI(0)), "operator/=(SafeInteger)");
      if ( SIDL::min() + SIDL::max() < 0 )
      {
         ASSERT_NOTHROW(SI(SIDL::max()) /= SI(-1), "operator/=(SafeInteger)");
         ASSERT_ANY_EXCEPTION(SI(SIDL::min()) /= SI(-1), "operator/=(SafeInteger)");
      }
      else if ( SIDL::min() + SIDL::max() > 0 )
      {
         ASSERT_NOTHROW(SI(SIDL::min()) /= SI(-1), "operator/=(SafeInteger)");
         ASSERT_ANY_EXCEPTION(SI(SIDL::max()) /= SI(-1), "operator/=(SafeInteger)");
      }
      else
      {
         ASSERT_NOTHROW(SI(SIDL::min()) /= SI(-1), "operator/=(SafeInteger)");
         ASSERT_NOTHROW(SI(SIDL::max()) /= SI(-1), "operator/=(SafeInteger)");
      }
   }

   void test_operator_add_assign()
   {
      ASSERT(((SI(6) += SI(3)) == SI(9)), "operator+=(SafeInteger)");
      ASSERT(((SI(6) += SI(-3)) == SI(3)), "operator+=(SafeInteger)");
      ASSERT(((SI(-6) += SI(3)) == SI(-3)), "operator+=(SafeInteger)");
      ASSERT(((SI(-6) += SI(-3)) == SI(-9)), "operator+=(SafeInteger)");
      ASSERT(((SI(0) += SI(3)) == SI(3)), "operator+=(SafeInteger)");
      ASSERT(((SI(6) += SI(0)) == SI(6)), "operator+=(SafeInteger)");
      ASSERT_NOTHROW(SI(SIDL::max()) += SI(0), "operator+=(SafeInteger)");
      ASSERT_ANY_EXCEPTION(SI(SIDL::max()) += SI(1), "operator+=(SafeInteger)");
      ASSERT_NOTHROW(SI(SIDL::min()) += SI(0), "operator+=(SafeInteger)");
      ASSERT_ANY_EXCEPTION(SI(SIDL::min()) += SI(-1), "operator+=(SafeInteger)");
   }

   void test_operator_subtract_assign()
   {
      ASSERT(((SI(6) -= SI(3)) == SI(3)), "operator-=(SafeInteger)");
      ASSERT(((SI(6) -= SI(-3)) == SI(9)), "operator-=(SafeInteger)");
      ASSERT(((SI(-6) -= SI(3)) == SI(-9)), "operator-=(SafeInteger)");
      ASSERT(((SI(-6) -= SI(-3)) == SI(-3)), "operator-=(SafeInteger)");
      ASSERT(((SI(0) -= SI(3)) == SI(-3)), "operator-=(SafeInteger)");
      ASSERT(((SI(6) -= SI(0)) == SI(6)), "operator-=(SafeInteger)");
      ASSERT_NOTHROW(SI(SIDL::max()) -= SI(0), "operator-=(SafeInteger)");
      ASSERT_ANY_EXCEPTION(SI(SIDL::max()) -= SI(-1), "operator-=(SafeInteger)");
      ASSERT_NOTHROW(SI(SIDL::min()) -= SI(0), "operator-=(SafeInteger)");
      ASSERT_ANY_EXCEPTION(SI(SIDL::min()) -= SI(1), "operator-=(SafeInteger)");
   }

   void test_operator_modulo_assign()
   {
      ASSERT(((SI(9) %= SI(7)) == SI(2)), "operator%=(SafeInteger)");
      ASSERT(((SI(-9) %= SI(7)) == SI(5)), "operator%=(SafeInteger)");
      ASSERT(((SI(0) %= SI(7)) == SI(0)), "operator%=(SafeInteger)");
      ASSERT_ANY_EXCEPTION(SI(9) %= SI(0), "operator%=(SafeInteger)");
      ASSERT_ANY_EXCEPTION(SI(9) %= SI(-1), "operator%=(SafeInteger)");
   }

   void test_operator_multiply()
   {
      ASSERT(((SI(2) * SI(3)) == SI(6)), "operator*(SafeInteger)");
      ASSERT(((SI(2) * SI(-3)) == SI(-6)), "operator*(SafeInteger)");
      ASSERT(((SI(-2) * SI(3)) == SI(-6)), "operator*(SafeInteger)");
      ASSERT(((SI(-2) * SI(-3)) == SI(6)), "operator*(SafeInteger)");
      ASSERT(((SI(0) * SI(3)) == SI(0)), "operator*(SafeInteger)");
      ASSERT(((SI(2) * SI(0)) == SI(0)), "operator*(SafeInteger)");
      ASSERT_NOTHROW(SI(SIDL::max()) * SI(1), "operator*(SafeInteger)")
      ASSERT_ANY_EXCEPTION(SI(SIDL::max()) * SI(2), "operator*(SafeInteger)")
      ASSERT_NOTHROW(SI(SIDL::min()) * SI(1), "operator*(SafeInteger)")
      ASSERT_ANY_EXCEPTION(SI(SIDL::min()) * SI(2), "operator*(SafeInteger)")
   }

   void test_operator_divide()
   {
      ASSERT(((SI(6) / SI(3)) == SI(2)), "operator/(SafeInteger)");
      ASSERT(((SI(6) / SI(-3)) == SI(-2)), "operator/(SafeInteger)");
      ASSERT(((SI(-6) / SI(3)) == SI(-2)), "operator/(SafeInteger)");
      ASSERT(((SI(-6) / SI(-3)) == SI(2)), "operator/(SafeInteger)");
      ASSERT(((SI(0) / SI(3)) == SI(0)), "operator/(SafeInteger)");
      ASSERT_ANY_EXCEPTION(((SI(6) / SI(0)) == SI(0)), "operator/(SafeInteger)");
      if ( SIDL::min() + SIDL::max() < 0 )
      {
         ASSERT_NOTHROW(SI(SIDL::max()) / SI(-1), "operator/(SafeInteger)");
         ASSERT_ANY_EXCEPTION(SI(SIDL::min()) / SI(-1), "operator/(SafeInteger)");
      }
      else if ( SIDL::min() + SIDL::max() > 0 )
      {
         ASSERT_NOTHROW(SI(SIDL::min()) / SI(-1), "operator/(SafeInteger)");
         ASSERT_ANY_EXCEPTION(SI(SIDL::max()) / SI(-1), "operator/(SafeInteger)");
      }
      else
      {
         ASSERT_NOTHROW(SI(SIDL::min()) / SI(-1), "operator/(SafeInteger)");
         ASSERT_NOTHROW(SI(SIDL::max()) / SI(-1), "operator/(SafeInteger)");
      }
   }

   void test_operator_add()
   {
      ASSERT(((SI(6) + SI(3)) == SI(9)), "operator+(SafeInteger)");
      ASSERT(((SI(6) + SI(-3)) == SI(3)), "operator+(SafeInteger)");
      ASSERT(((SI(-6) + SI(3)) == SI(-3)), "operator+(SafeInteger)");
      ASSERT(((SI(-6) + SI(-3)) == SI(-9)), "operator+(SafeInteger)");
      ASSERT(((SI(0) + SI(3)) == SI(3)), "operator+(SafeInteger)");
      ASSERT(((SI(6) + SI(0)) == SI(6)), "operator+(SafeInteger)");
      ASSERT_NOTHROW(SI(SIDL::max()) + SI(0), "operator+(SafeInteger)");
      ASSERT_ANY_EXCEPTION(SI(SIDL::max()) + SI(1), "operator+(SafeInteger)");
      ASSERT_NOTHROW(SI(SIDL::min()) + SI(0), "operator+(SafeInteger)");
      ASSERT_ANY_EXCEPTION(SI(SIDL::min()) + SI(-1), "operator+(SafeInteger)");
   }

   void test_operator_subtract()
   {
      ASSERT(((SI(6) - SI(3)) == SI(3)), "operator-(SafeInteger)");
      ASSERT(((SI(6) - SI(-3)) == SI(9)), "operator-(SafeInteger)");
      ASSERT(((SI(-6) - SI(3)) == SI(-9)), "operator-(SafeInteger)");
      ASSERT(((SI(-6) - SI(-3)) == SI(-3)), "operator-(SafeInteger)");
      ASSERT(((SI(0) - SI(3)) == SI(-3)), "operator-(SafeInteger)");
      ASSERT(((SI(6) - SI(0)) == SI(6)), "operator-(SafeInteger)");
      ASSERT_NOTHROW(SI(SIDL::max()) - SI(0), "operator-(SafeInteger)");
      ASSERT_ANY_EXCEPTION(SI(SIDL::max()) - SI(-1), "operator-(SafeInteger)");
      ASSERT_NOTHROW(SI(SIDL::min()) - SI(0), "operator-(SafeInteger)");
      ASSERT_ANY_EXCEPTION(SI(SIDL::min()) - SI(1), "operator-(SafeInteger)");
   }

   void test_operator_modulo()
   {
      ASSERT(((SI(9) % SI(7)) == SI(2)), "operator%(SafeInteger)");
      ASSERT(((SI(-9) % SI(7)) == SI(5)), "operator%(SafeInteger)");
      ASSERT(((SI(0) % SI(7)) == SI(0)), "operator%(SafeInteger)");
      ASSERT_ANY_EXCEPTION(((SI(9) % SI(0)) == SI(0)), "operator%(SafeInteger)");
      ASSERT_ANY_EXCEPTION(SI(9) % SI(0), "operator%(SafeInteger)");
      ASSERT_ANY_EXCEPTION(SI(9) % SI(-1), "operator%(SafeInteger)");
   }
}

